#######################################################################################################################
# Author: Maxime Schmitt
# Mail: maxime.schmitt@telecom-bretagne.eu
# Overview: Module that combines all the other modules and provides a CLI.
#######################################################################################################################

import sys
import click
import os
import subprocess
import shutil
import time
import glob

sys.path.append('.')
import configManager
import bootstrapEB
import buildSwSets

import pkg_resources

#######################################################################################################################
# The resif group. Defines the name of the command. It is the "main" group.
@click.group(invoke_without_command=True)
@click.pass_context
@click.option('--version', flag_value=True, help='Return the version of this script.')
def resif(ctx, version):
    """
    RESIF commandline interface.

    Choose the sub-command you want to execute.
    """
    if ctx.invoked_subcommand is None:
        if version:
            sys.stdout.write("This is RESIF version " + pkg_resources.require("resif")[0].version + "\n")
        else:
            subprocess.check_call(['resif', '--help'])
        
#######################################################################################################################


#######################################################################################################################
# The init, update and wipe subcommands, to initialize, update and reset the environment concerning RESIF

# Initialize the necessary directories
@resif.command(short_help='Initialize the git repository in the srcpath.')
@click.option('--git-architecture', 'git_architecture', envvar='RESIF_GIT_ARCHITECTURE', help='Defines an alternative git repository URL or absolute path to get the architecture from.')
@click.option('--srcpath', 'srcpath', envvar='RESIF_SRCPATH', help='Defines an alternative path to put the sources in.')
@click.option('--overwrite', 'overwrite', flag_value=True, envvar='RESIF_OVERWRITE', help='Set this flag if you want to overwrite any existing directories in the srcpath.')
def init(**kwargs):
    config = configManager.generateInitConfig(kwargs)
    if not os.path.isdir(config["srcpath"]) or config["overwrite"]:
        if config["overwrite"]:
            shutil.rmtree(config["srcpath"], True)
        subprocess.check_call(['git', 'clone', config['git_architecture'], config['srcpath']])
    else:
        sys.stdout.write("A repository already exist at your srcpath: " + config["srcpath"] +"\nPlease use the --overwrite flag if you want to overwrite this repository.\n" + "\033[93m" + "WARNING: This will remove everything at " + config["srcpath"] + "\033[0m\n")
        exit(50)

@resif.command(short_help='Update the git repository in the srcpath.')
@click.option('--srcpath', 'srcpath', envvar='RESIF_SRCPATH', help='Defines an alternative path to the repository.')
def update(**kwargs):
    config = configManager.generateUpdateConfig(kwargs)
    os.chdir(config['srcpath'])
    subprocess.check_call(['git', 'pull'])

@resif.command(short_help='Wipe all data in the srcpath.')
@click.option('--srcpath', 'srcpath', envvar='RESIF_SRCPATH', help='Defines an alternative path to the repository.')
@click.confirmation_option(prompt='You are going to remove everything in <srcpath> (default=$HOME/.resif/src), are you sure you want to continue ?', help='Use to not prompt confirmation message. (Check what you are trying to do before !)')
def wipe(**kwargs):
    config = configManager.generateWipeConfig(kwargs)
    try:
        shutil.rmtree(config['srcpath'])
    except OSError:
        sys.stdout.write("Nothing to remove at " + config['srcpath'] + "\n")

#######################################################################################################################


#######################################################################################################################
# The count and show subcommands which help getting informations such as
# "how many and which software have been compiled using X toolchain?"
# These functions need one of the "LOADME" files to be loaded.
# (At the very least, they require the module path to be set correctly to use the EasyBuild install and the RESIF_ROOTINSTALL variable to be set to the root of the EasyBuild install.)

# Show the list of softwares which easyconfig file contains the "content" string
@resif.command(short_help="Show all the easyconfigs which name contains the given pattern.")
@click.option('--rootinstall', envvar='RESIF_ROOTINSTALL', help='Path to the root of the EasyBuild installation (contains the various software sets deployed and the EasyBuild files).')
@click.option('--mns', envvar='EASYBUILD_MODULE_NAMING_SCHEME', type=click.Choice(['EasyBuildMNS', 'E', 'HierarchicalMNS', 'H', 'ThematicMNS', 'T']), help='Module Naming Scheme to be used.')
@click.option('--show-path', 'show_path', flag_value=True, envvar='RESIF_SHOW_PATH', help='Make the command to show the full path to the files listed.')
@click.argument('content')
def search(**kwargs):
    """
    [CONTENT] TEXT                  Text to look for in the names of
                                    the installed softwares.
    """
    try:
        if kwargs['mns'] == 'ThematicMNS' or kwargs['mns'] == 'T':
            easybuild_module = "base/EasyBuild/install-" + configManager.getEasyBuildVersion(os.path.expandvars(os.path.abspath(os.environ['RESIF_ROOTINSTALL'])))
        else:
            easybuild_module = 'EasyBuild/install-' + configManager.getEasyBuildVersion(os.path.expandvars(os.path.abspath(os.environ['RESIF_ROOTINSTALL'])))
    except Exception:
        sys.stdout.write("Please set the RESIF_ROOTINSTALL environment variable or the --rootinstall option to the root of the EasyBuild installation.\n")
        exit(1)
    process = subprocess.Popen('/bin/bash', stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    process.stdin.write('module load ' + easybuild_module + '\n')
    if kwargs['show_path']:
        process.stdin.write('eb -S ' + kwargs['content'] + " | grep '^[^=]' | sed 's#^ \* ##'\n")
    else:
        process.stdin.write('eb -S ' + kwargs['content'] + " | grep '^ \* ' | sed 's#.*/##'\n")
    process.stdin.write('echo $?\n')
    out = ""
    while True:
        out = process.stdout.readline()
        try:
            i = int(out)
        except ValueError:
            i = -1
        if i < 0:
            sys.stdout.write(out)
        else:
            break

#Count the number of easyconfig files that contain the given pattern in their name.
@resif.command(short_help="Count the number of easyconfig files that contain the given pattern in their name.")
@click.option('--rootinstall', envvar='RESIF_ROOTINSTALL', help='Path to the root of the EasyBuild installation (contains the various software sets deployed and the EasyBuild files).')
@click.option('--mns', envvar='EASYBUILD_MODULE_NAMING_SCHEME', type=click.Choice(['EasyBuildMNS', 'E', 'HierarchicalMNS', 'H', 'ThematicMNS', 'T']), help='Module Naming Scheme to be used.')
@click.argument('content')
def count(**kwargs):
    """
    \b
    [CONTENT] TEXT                  Text to look for in the names of
                                    the easyconfig files.
    """
    try:
        if kwargs['mns'] == 'ThematicMNS' or kwargs['mns'] == 'T':
            easybuild_module = "base/EasyBuild/install-" + configManager.getEasyBuildVersion(os.path.expandvars(os.path.abspath(kwargs['rootinstall'])))
        else:
            easybuild_module = 'EasyBuild/install-' + configManager.getEasyBuildVersion(os.path.expandvars(os.path.abspath(kwargs['rootinstall'])))
    except Exception:
        sys.stdout.write("Please set the RESIF_ROOTINSTALL environment variable or the --rootinstall option to the root of the EasyBuild installation.\n")
        exit(1)
    process = subprocess.Popen('/bin/bash', stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    process.stdin.write('module load ' + easybuild_module + '\n')
    process.stdin.write('eb -S ' + kwargs['content'] + ' | grep "^ \* " | wc -l' + '\n')
    sys.stdout.write(process.stdout.readline())


# Print the build duration of all the softwares in the given software set.
@resif.command(short_help="Print the build times of all the installed software of a given software set.")
@click.option('--rootinstall', envvar='RESIF_ROOTINSTALL', help='Path to the root of the EasyBuild installation (contains the various software sets deployed and the EasyBuild files).')
@click.option('--seconds', flag_value=True, envvar='RESIF_SECONDS', help='Set this flag if you want the software build time output to be in seconds (not formated).')
@click.argument('swset')
def buildTimeSwSet(**kwargs):
    """
    [SWSET] TEXT        Software set to consider.
    """
    files = glob.glob(kwargs['rootinstall']+'/'+kwargs['swset']+'/software/*/*/*/easybuild/*log')
    coreTime = 0
    if files != []:
        for logfile in files:
            software, softwareDuration = buildSwSets.getSoftwareBuildTimes(logfile)
            coreTime += softwareDuration
            if kwargs['seconds']:
                sys.stdout.write(software + "\t" + str(softwareDuration) + "\n")
            else:
                m, s = divmod(softwareDuration, 60)
                h, m = divmod(m, 60)
                softwareDurationFormated = "%d:%d:%d" % (h, m, s)
                sys.stdout.write(software + "\t" + softwareDurationFormated + "\n")
        # Output the build time of the whole software set.
        if kwargs['seconds']:
            sys.stdout.write("core software set\t" + str(coreTime) + "\n")
        else:
            m, s = divmod(coreTime, 60)
            h, m = divmod(m, 60)
            coreTimeStr = "%d:%d:%d" % (h, m, s)
            sys.stdout.write("core software set\t" + coreTimeStr + "\n")
    else:
        sys.stdout.write("No software found.\n")
        exit(80)


# Print the build durations of all the versions of the given software if it is found.
@resif.command(short_help="Print the build time of of a given installed software from a specific software set.")
@click.option('--rootinstall', envvar='RESIF_ROOTINSTALL', help='Path to the root of the EasyBuild installation (contains the various software sets deployed and the EasyBuild files).')
@click.option('--seconds', flag_value=True, envvar='RESIF_SECONDS', help='Set this flag if you want the software build time output to be in seconds (not formated).')
@click.argument('swset')
@click.argument('software')
def buildTimeSoftware(**kwargs):
    """
    \b
    [SWSET] TEXT        Software set to consider.
    [SOFTWARE] TEXT     Software name to consider.
    """
    softwareFound = False
    files = glob.glob(kwargs['rootinstall']+'/'+kwargs['swset']+'/software/*/*/*/easybuild/*log')
    for logfile in files:
        software, softwareDuration = buildSwSets.getSoftwareBuildTimes(logfile)
        if software.lower() == kwargs['software'].lower():
            softwareFound = True
            if kwargs['seconds']:
                sys.stdout.write(software + "\t" + str(softwareDuration) + "\n")
            else:
                m, s = divmod(softwareDuration, 60)
                h, m = divmod(m, 60)
                softwareDurationFormated = "%d:%d:%d" % (h, m, s)
                sys.stdout.write(software + "\t" + softwareDurationFormated + "\n")
            # We do not exit() since there may be multiple version of the same software.
    if not softwareFound:
        sys.stdout.write("The software you asked for has not been found. This software is either not installed or in another software set.\n")
        exit(90)

#######################################################################################################################


#######################################################################################################################
# The subcommands bootstrap, build and cleaninstall.

# Make a new install of EasyBuild.
@resif.command(short_help='Deploy a fresh EasyBuild install.')
# Path to the source directory
@click.option('--srcpath', envvar='RESIF_SRCPATH', help='Source path to the RESIF directory.')
# Configuration files provisioning
@click.option('--configfile', envvar='RESIF_CONFIGFILE', help='Specify path to a config file to load.')
# git variables
@click.option('--gh-ebuser', 'gh_ebuser', envvar='RESIF_GH_EBUSER', help='Specify a GitHub user that has the EasyBuild repositories you want to use instead of the one provided by ULHPC.')
@click.option('--git-ebframework', 'git_ebframework', envvar='RESIF_GIT_EBFRAMEWORK', help='URL or absolute path to EasyBuild framework Git repository.')
@click.option('--git-ebblocks', 'git_ebblocks', envvar='RESIF_GIT_EBBLOCKS', help='URL or absolute path for EasyBuild easyblocks Git repository.')
@click.option('--git-ebconfigs', 'git_ebconfigs', envvar='RESIF_GIT_EBCONFIGS', help='URL or absolute path for EasyBuild easyconfigs Git repository.')
@click.option('--branch-ebframework', 'branch_ebframework', envvar='RESIF_BRANCH_EBFRAMEWORK', help='Git branch for EasyBuild framework.')
@click.option('--branch-ebblocks', 'branch_ebblocks', envvar='RESIF_BRANCH_EBBLOCKS', help='Git branch for EasyBuild easyblocks.')
@click.option('--branch-ebconfigs', 'branch_ebconfigs', envvar='RESIF_BRANCH_EBCONFIGS', help='Git branch for EasyBuild easyconfigs.')
@click.option('--subtrees', flag_value=True, envvar='RESIF_SUBTREES', help='Set this flag if you want to allow usage of subtrees from modules repository as sources for the EasyBuild components. (Backward compability option)')
# Bootstrap application variables
@click.option('--apps-root', 'apps_root', envvar='RESIF_APPS_ROOT', help='Path to the root directory for apps (contains all the architecture correspondig to RESIF).')
@click.option('--production', 'branch', flag_value='production', envvar='RESIF_BRANCH', help='Set this variable if you want to work with the production branch of the RESIF repository. (By default, work with the production branch). Do not use with --devel or --branch !')
@click.option('--devel', 'branch', flag_value='devel', envvar='RESIF_BRANCH', help='Set this variable if you want to work with the devel branch of the RESIF repository. (By default, work with the production branch). Do not use with --production or --branch !')
@click.option('--branch', 'branch', envvar='RESIF_BRANCH', help='Set this variable if you want to work with the devel branch of the RESIF repository. (By default, work with the production branch). Do not use with --production or --branch !')
@click.option('--release', envvar='RESIF_RELEASE', help='Release tag or commit of the RESIF repository to deploy.')
@click.option('--releasedir', envvar='RESIF_RELEASEDIR', help='Directory in which to install the release (Relative path from the <apps-root>). Default: <branch>/v<release>-<date>')
@click.option('--rootinstall', envvar='RESIF_ROOTINSTALL', help='Path to the root of the EasyBuild installation (contains the various software sets deployed and the EasyBuild files). Default: <apps-root>/<releasedir>')
@click.option('--append-modulepath', 'append_modulepath', envvar='RESIF_APPEND_MODULEPATH', help='Paths to append to the modulepath in the LOADME files.')
@click.option('--prepend-modulepath', 'prepend_modulepath', envvar='RESIF_PREPEND_MODULEPATH', help='Paths to prepend to the modulepath in the LOADME files.')
# Module Naming Scheme choice
@click.option('--mns', envvar='EASYBUILD_MODULE_NAMING_SCHEME', type=click.Choice(['EasyBuildMNS', 'E', 'HierarchicalMNS', 'H', 'ThematicMNS', 'T']), help='Module Naming Scheme to be used.')
@click.option('--overwrite', 'overwrite', flag_value=True, envvar='RESIF_OVERWRITE', help='Set this flag if you want to overwrite any existing previous installation at --apps-root.')
def bootstrap(**kwargs):
    # Generate the configuration for the bootstrap.
    config = configManager.generateBootstrapConfig(kwargs)
    # Bootstrap EasyBuild.
    if not os.path.isdir(config["rootinstall"]) or config["overwrite"]:
        if config["overwrite"]:
            shutil.rmtree(config["rootinstall"], True)
        click.echo("Bootstrapping EasyBuild.")
        bootstrapEB.bootstrap(config)
        click.echo("Bootstrapping ended successfully.")
        sourcemePath = os.path.join(config['rootinstall'], 'LOADME-'+os.path.basename(config['rootinstall'])+'.sh')
        click.echo("\nTo start using this installation, source the following file:\n" + sourcemePath)
    else:
        sys.stdout.write("An installation is already present at your rootinstall: " + config["rootinstall"] + "\nPlease use the --overwrite flag if you want to overwrite this installation.\n" + "\033[93m" + "WARNING: This will remove everything at " + config["rootinstall"] + "\033[0m\n")
        exit(50)


# Build a (or multiple) software set(s) (Adding new software to an existing EasyBuild install.)
@resif.command(short_help='Deploy software sets on an existing installatation.')
# Path to the source directory
@click.option('--srcpath', envvar='RESIF_SRCPATH', help='Source path to the RESIF directory.')
# Configuration files provisioning
@click.option('--configfile', envvar='RESIF_CONFIGFILE', help='Specify path to a config file to load.')
# Software building variables
@click.option('--rootinstall', envvar='RESIF_ROOTINSTALL', help='Path to the root of an EasyBuild installation (contains the various software sets deployed and the EasyBuild files). Softwares will be installed in <rootinstall>/<swset>/modules')
@click.option('--production', 'branch', flag_value='production', envvar='RESIF_BRANCH', help='Set this variable if you want to work with the production branch of the RESIF repository. (By default, work with the production branch). Do not use with --devel or --branch !')
@click.option('--devel', 'branch', flag_value='devel', envvar='RESIF_BRANCH', help='Set this variable if you want to work with the devel branch of the RESIF repository. (By default, work with the production branch). Do not use with --production or --branch !')
@click.option('--branch', 'branch', envvar='RESIF_BRANCH', help='Set this variable if you want to work with the devel branch of the RESIF repository. (By default, work with the production branch). Do not use with --production or --branch !')
@click.option('--release', envvar='RESIF_RELEASE', help='Release tag or commit of the RESIF repository to use.')
@click.option('--force', 'force', flag_value=True, envvar='RESIF_FORCE', help='Set this flag if you want to force build even if release or branch of the existing stack doesn\'t match given ones.')
@click.option('--installdir', 'installdir', envvar='RESIF_INSTALLDIR', help="Use if you don't want to deploy the software inside the <rootinstall>. Softwares will then be deployed in <installdir>/<swset>/modules")
@click.option('--eb-sourcepath', 'eb_sourcepath', envvar='EASYBUILD_SOURCEPATH', help='EasyBuild sourcepath.')
@click.option('--eb-buildpath', 'eb_buildpath', envvar='EASYBUILD_BUILDPATH', help='EasyBuild buildpath.')
@click.option('--eb-repository', 'eb_repository', envvar='EASYBUILD_REPOSITORY', help='EasyBuild repository type for successfully installed easyconfig files.')
@click.option('--eb-repositorypath', 'eb_repositorypath', envvar='EASYBUILD_REPOSITORYPATH', help='EasyBuild path to the repository for successuflly installed easyconfig files.')
@click.option('--eb-options', 'eb_options', envvar='RESIF_EB_OPTIONS', help='Any command line options to pass to EasyBuild for the build.')
@click.option('--buildmode', envvar='RESIF_BUILDMODE', type=click.Choice(['local', 'job']), help='Mode to build the software: either building locally or in a job.')
@click.option('--mns', envvar='EASYBUILD_MODULE_NAMING_SCHEME', type=click.Choice(['EasyBuildMNS', 'E', 'HierarchicalMNS', 'H', 'ThematicMNS', 'T']), help='Module Naming Scheme to be used.')
@click.option('--out-place', 'out_place', flag_value=True, envvar='RESIF_OUT_PLACE', help='Set this option if you want all the files (sources, build, repository) to be put outside the rootinstall (in an associated subdirectory in $HOME/.resif).')
@click.option('--swsets-config', 'swsets_config', envvar='RESIF_SWSETS_CONFIG', help='Path to a file defining the software sets.')
@click.argument('swsets', nargs=-1)
def build(**kwargs):
    """
    \b
    [SWSETS] TEXT...                Software sets to deploy.
    """
    # Generate the configuration for the bootstrap
    config = configManager.generateBuildConfig(kwargs)
    # Build the software sets.
    click.echo("Building the software sets.")
    config['easybuild_module'] = configManager.getEasyBuildModule(config)
    start = time.time()
    buildSwSets.build(config)
    end = time.time()
    duration = end - start
    m, s = divmod(duration, 60)
    h, m = divmod(m, 60)
    durationFormated = "%dh%dm%ds" % (h, m, s)
    click.echo("Software sets successfully built. The build duration was of " + durationFormated)
    # We return a list of modulepaths and modules to load to start using the new software sets
    modulepaths = ""
    swsetsModulesList = ""
    for swset in config['swsets']:
        if swset != 'core':
            swsetsModulesList += "base/swsets/" + swset + " "
        if 'installdir' in config:
            modulepaths += os.path.join(os.path.join(os.path.join(config['installdir'], swset), 'modules'), 'all') + ':'
        elif config['out_place']:
            try:
                modulepaths += os.path.join(os.environ['EASYBUILD_INSTALLPATH'], 'modules', 'all') + ':'
            except Exception:
                modulepaths += os.path.join(os.path.expandvars('$HOME'), '.local', 'easybuild', 'modules', 'all') + ':'
        else:
            modulepaths += os.path.join(os.path.join(os.path.join(config['rootinstall'], swset), 'modules'), 'all') + ':'
    if swsetsModulesList != "" and not config['out_place']:
        sys.stdout.write("\nTo make the software sets available, add the following paths to your MODULEPATH environment variable:\n" + modulepaths + "\nOr load the following modules:\n" + swsetsModulesList + "\n")
    elif not 'installdir' in config and config['out_place']:
        sys.stdout.write('\nTo make the software sets available, add the following paths to your MODULEPATH environment variable:\n' + modulepaths + '\n')
    else:
        sys.stdout.write("\nTo make the software sets available, add the following paths to your MODULEPATH environment variable:\n" + modulepaths + "\n")


# Full install (Correspond to making a new release)
@resif.command(short_help='Deploy a full environment: bootstrap EasyBuild and use it to install the software sets.')
# Path to the source directory
@click.option('--srcpath', envvar='RESIF_SRCPATH', help='Source path to the RESIF directory.')
# Configuration files provisioning
@click.option('--configfile', envvar='RESIF_CONFIGFILE', help='Specify path to a config file to load.')
# git variables
@click.option('--gh-ebuser', 'gh_ebuser', envvar='RESIF_GH_EBUSER', help='Specify a GitHub user that has the EasyBuild repositories you want to use instead of the one provided by ULHPC.')
@click.option('--git-ebframework', 'git_ebframework', envvar='RESIF_GIT_EBFRAMEWORK', help='URL or absolute path to EasyBuild framework Git repository.')
@click.option('--git-ebblocks', 'git_ebblocks', envvar='RESIF_GIT_EBBLOCKS', help='URL or absolute path for EasyBuild easyblocks Git repository.')
@click.option('--git-ebconfigs', 'git_ebconfigs', envvar='RESIF_GIT_EBCONFIGS', help='URL or absolute path for EasyBuild easyconfigs Git repository.')
@click.option('--branch-ebframework', 'branch_ebframework', envvar='RESIF_BRANCH_EBFRAMEWORK', help='Git branch for EasyBuild framework.')
@click.option('--branch-ebblocks', 'branch_ebblocks', envvar='RESIF_BRANCH_EBBLOCKS', help='Git branch for EasyBuild easyblocks.')
@click.option('--branch-ebconfigs', 'branch_ebconfigs', envvar='RESIF_BRANCH_EBCONFIGS', help='Git branch for EasyBuild easyconfigs.')
@click.option('--subtrees', flag_value=True, envvar='RESIF_SUBTREES', help='Set this flag if you want to allow usage of subtrees from modules repository as sources for the EasyBuild components. (Backward compability option)')
# Bootstrap application variables
@click.option('--apps-root', 'apps_root', envvar='RESIF_APPS_ROOT', help='Path to the root directory for apps (contains all the architecture correspondig to RESIF).')
@click.option('--production', 'branch', flag_value='production', envvar='RESIF_BRANCH', help='Set this variable if you want to work with the production branch of the RESIF repository. (By default, work with the production branch). Do not use with --devel or --branch !')
@click.option('--devel', 'branch', flag_value='devel', envvar='RESIF_BRANCH', help='Set this variable if you want to work with the devel branch of the RESIF repository. (By default, work with the production branch). Do not use with --production or --branch !')
@click.option('--branch', 'branch', envvar='RESIF_BRANCH', help='Set this variable if you want to work with the devel branch of the RESIF repository. (By default, work with the production branch). Do not use with --production or --branch !')
@click.option('--release', envvar='RESIF_RELEASE', help='Release tag or commit of the RESIF repository to deploy.')
@click.option('--releasedir', envvar='RESIF_RELEASEDIR', help='Directory in which to install the release (Relative path from the <apps-root>). Default: <branch>/v<release>-<date>')
@click.option('--rootinstall', envvar='RESIF_ROOTINSTALL', help='Path to the root of the EasyBuild installation (contains the various software sets deployed and the EasyBuild files). Default: <apps-root>/<releasedir>')
@click.option('--eb-sourcepath', 'eb_sourcepath', envvar='EASYBUILD_SOURCEPATH', help='EasyBuild sourcepath.')
@click.option('--eb-buildpath', 'eb_buildpath', envvar='EASYBUILD_BUILDPATH', help='EasyBuild buildpath.')
@click.option('--eb-repository', 'eb_repository', envvar='EASYBUILD_REPOSITORY', help='EasyBuild repository type for successfully installed easyconfig files.')
@click.option('--eb-repositorypath', 'eb_repositorypath', envvar='EASYBUILD_REPOSITORYPATH', help='EasyBuild path to the repository for successuflly installed easyconfig files.')
@click.option('--eb-options', 'eb_options', envvar='RESIF_EB_OPTIONS', help='Any command line options to pass to EasyBuild for the build.')
@click.option('--append-modulepath', 'append_modulepath', envvar='RESIF_APPEND_MODULEPATH', help='Paths to append to the modulepath in the LOADME files. Also won\'t reinstall software found at these paths.')
@click.option('--prepend-modulepath', 'prepend_modulepath', envvar='RESIF_PREPEND_MODULEPATH', help='Paths to prepend to the modulepath in the LOADME files. Also won\'t reinstall software found at these paths.')
# Module Naming Scheme choice
@click.option('--mns', envvar='EASYBUILD_MODULE_NAMING_SCHEME', type=click.Choice(['EasyBuildMNS', 'E', 'HierarchicalMNS', 'H', 'ThematicMNS', 'T']), help='Module Naming Scheme to be used.')
# Software building variables
@click.option('--buildmode', envvar='RESIF_BUILDMODE', type=click.Choice(['local', 'job']), help='Mode to build the software: either building locally or in a job.')
@click.option('--swsets-config', 'swsets_config', envvar='RESIF_SWSETS_CONFIG', help='Path to a file defining the software sets.')
@click.option('--overwrite', 'overwrite', flag_value=True, envvar='RESIF_OVERWRITE', help='Set this flag if you want to overwrite any existing previous installation at --apps-root.')
@click.argument('swsets', nargs=-1)
def cleaninstall(**kwargs):
    """
    [SWSETS] TEXT...                Software set to deploy.
    """
    # Generate the configuration for the full installation
    config = configManager.generateCleaninstallConfig(kwargs)
    click.echo("Starting full installation.")
    # Bootstrap EasyBuild.
    if not os.path.isdir(config["rootinstall"]) or config["overwrite"]:
        if config["overwrite"]:
            shutil.rmtree(config["rootinstall"], True)
        click.echo("Bootstrapping EasyBuild.")
        modulePath = bootstrapEB.bootstrap(config)
    else:
        sys.stdout.write("An installation is already present at your rootinstall: " + config["rootinstall"] + "\nPlease use the --overwrite flag if you want to overwrite this installation.\n" + "\033[93m" + "WARNING: This will remove everything at " + config["rootinstall"] + "\033[0m\n")
        exit(50)
    click.echo("Bootstrapping ended successfully.")
    # Build the software sets.
    click.echo("Building the software sets.")
    # Setting the correct MODULEPATH and EasyBuild variables.
    # (Necessary for the behavior to not be modified by external environment variables)
    if config["module_cmd"] == "lmod":
        os.environ["EASYBUILD_MODULES_TOOL"] = "Lmod"
        subprocess.Popen(["bash", "-c", "module unuse $MODULEPATH"])
    os.environ['MODULEPATH'] = modulePath
    configManager.setEasyBuildVariables(config)
    config['easybuild_module'] = configManager.getEasyBuildModule(config)
    # To avoid environment variable interference, we will unset EASYBUILD_INSTALLPATH when building knowing we're doing a cleaninstall
    config['iscleaninstall'] = True
    start = time.time()
    buildSwSets.build(config)
    end = time.time()
    duration = end - start
    m, s = divmod(duration, 60)
    h, m = divmod(m, 60)
    durationFormated = "%dh%dm%ds" % (h, m, s)
    click.echo("Software sets successfully built. The build duration was of " + durationFormated)
    click.echo("Full installation ended successfully.")
    sourcemePath = os.path.join(config['rootinstall'], 'LOADME-'+os.path.basename(config['rootinstall'])+'.sh')
    click.echo("\nTo start using this installation, source the following file:\n" + sourcemePath)


#######################################################################################################################


#######################################################################################################################
# Author: Maxime Schmitt
# Mail: maxime.schmitt@telecom-bretagne.eu
# Overview: Module that take care of managing the configuration files for the various functions of the RESIF suite.
#######################################################################################################################

import os
import sys
import yaml
import time
import re
import subprocess
from git import Repo

#######################################################################################################################
# These functions are wrapper to make the code more readable.
# They shouldn't do much more than calling the bellow functions.

def generateInitConfig(hashTable):
    if hashTable['git_architecture'] == None:
        hashTable['git_architecture'] = 'https://github.com/ULHPC/modules.git'

    if hashTable['srcpath'] == None:
        hashTable['srcpath'] = os.path.join(os.path.join('$HOME', '.resif'), 'src')

    encoder(hashTable, 'utf8')
    configExpandVars(hashTable)

    return hashTable

def generateUpdateConfig(hashTable):
    if hashTable['srcpath'] == None:
        hashTable['srcpath'] = os.path.join(os.path.join('$HOME', '.resif'), 'src')

    encoder(hashTable, 'utf8')
    configExpandVars(hashTable)

    return hashTable

def generateWipeConfig(hashTable):
    if not 'srcpath' in hashTable == None:
        hashTable['srcpath'] = os.path.join(os.path.join('$HOME', '.resif'), 'src')

    encoder(hashTable, 'utf8')
    configExpandVars(hashTable)

    return hashTable

def generateBootstrapConfig(hashTable):
    config = generateCommonConfig(hashTable)

    # Generate some missing configuration informations.
    if not 'releasedir' in config:
        generateReleasedir(config)

    if not 'rootinstall' in config:
        generateRootinstall(config)

    # Expanding apps_root and rootinstall to an absolute path.
    config['apps_root'] = os.path.abspath(config['apps_root'])
    config['rootinstall'] = os.path.abspath(config['rootinstall'])

    # Determining the branches to use for the EasyBuild git repositories
    resolveEBbranches(config)

    # Create the configuration file for EasyBuild.
    easybuildConfigfileCreator(config)
    
    return config

def generateBuildConfig(hashTable):
    config = generateCommonConfig(hashTable)

    # if rootinstall is given, expand it to obtain an abosulte path.
    if 'rootinstall' in config:
        config['rootinstall'] = os.path.abspath(config['rootinstall'])

    # If no swset is given, the script stops
    if not 'swsets' in config:
        sys.exit('Please specify at least one software set you want to build.\n')

    # Check that the release or branch (if any) corresponds to the actual release or branch of the software stack
    if not isValidRelease(config) and not config["force"]:
        sys.stdout.write("The release or branch given doesn't match the release or branch of the existing build.\n")
        exit(30)

    # Create the configuration file for EasyBuild.
    easybuildConfigfileCreator(config)

    return config


def generateCleaninstallConfig(hashTable):
    config = generateCommonConfig(hashTable)

    # Generate some missing configuration informations.
    if not 'releasedir' in config:
        generateReleasedir(config)

    if not 'rootinstall' in config:
        generateRootinstall(config)

    # Expanding apps_root and rootinstall to an absolute path.
    config['apps_root'] = os.path.abspath(config['apps_root'])
    config['rootinstall'] = os.path.abspath(config['rootinstall'])

    # Determining the branches to use for the EasyBuild git repositories
    resolveEBbranches(config)

    # Create the configuration file for EasyBuild.
    easybuildConfigfileCreator(config)

    # If no swset is given, the script stops
    if not 'swsets' in config:
        sys.exit('Please specify at least one software set you want to build.\n')

    # For the build to install everything inside the rootinstall, we set out_place to False
    config['out_place'] = False

    return config

#######################################################################################################################


#######################################################################################################################
# Utilities functions (move to a separate module ?)

# Encode all the field of a dict to a given encoding (Assume all the fields are strings or tuples of strings).
def encoder(hashTable, encoding):
    for k,v in hashTable.iteritems():
        if isinstance(v, tuple):
            x = ()
            for s in v:
                if isinstance(s, basestring):
                    x += (s.encode(encoding),)
            hashTable[k] = x
        else:
            if v != None and isinstance(v, basestring):
                hashTable[k] = v.encode(encoding)
    return hashTable


# Merge two dict into one (the first one).
def configMerger(default, user):
    if isinstance(default,dict) and isinstance(user,dict):
        for k,v in user.iteritems():
            if v != None and v != ():
                default[k] = v


# Take the path to a YAML file that contains the configuration and return a dict containing the associated config.
def configParser(configFile):
    config = yaml.load(configFile)

    for k in ['swsets']:
        if k in config:
            tup = ()
            for v in config[k]:
                tup += (v,)
            config[k] = tup
    
    return config


# Take a dict and expands any environment variable contained in its fields.
def configExpandVars(hashTable):
    for k,v in hashTable.iteritems():
        if isinstance(v, tuple):
            tup = ()
            for s in v:
                if isinstance(s, basestring):
                    tup += (os.path.expandvars(s),)
            hashTable[k] = tup
        else:
            if v != None and isinstance(v, basestring):
                hashTable[k] = os.path.expandvars(v)

# Check if a command exists in the current environment
def cmd_exists(cmd):
    return subprocess.call("type " + cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0

#######################################################################################################################


#######################################################################################################################
# This functions do the actual work in building the configuration.

# Generate the configuration for all the scenarios until differentiation is needed.
def generateCommonConfig(hashTable):
    # We load all configuration made by the user (through a configuration file or the options/environment variables)
    if hashTable['configfile'] != None:
        with open(os.path.join(os.getcwd(),os.path.expandvars(hashTable['configfile'])), "r") as f:
            userConfig = configParser(f)
        configMerger(userConfig, encoder(hashTable, 'utf8'))
    else:
        userConfig = {} # Trick to get rid of all the 'None' values
        configMerger(userConfig, encoder(hashTable, 'utf8'))

    # Environment variables are already managed through Click (envvar entry in the options)

    # If the the "srcpath" key has not been defined in the dict through one of the previous ways, we assume that it has its default value
    if not "srcpath" in userConfig:
        userConfig['srcpath'] = os.path.join(os.path.expandvars('$HOME'), '.resif/src')

    try:
        repo = Repo(userConfig['srcpath'])
    except Exception:
        sys.exit("Invalid git repository at " + userConfig['srcpath'])

    # If a a branch or a release has been given, we change the state of the repository accordingly, if not, we use the production branch
    if 'release' in userConfig or 'branch' in userConfig:
        if 'release' in userConfig:
            tree = repo.commit(userConfig['release']).tree
        else:
            tree = repo.heads[userConfig['branch']].commit.tree
    else:
        tree = repo.commit('HEAD').tree

    # We load the default config file and use it to complete the configuration given by the user
    defaultConfigFile = tree['config/config.yaml'].data_stream.read()
    config = configParser(defaultConfigFile)
    configMerger(config, userConfig)

    # Expanding all the environment variables (if any)
    configExpandVars(config)

    # Make srcpath an absolute path
    config['srcpath'] = os.path.abspath(config['srcpath'])

    # Replace short names for the MNS with the real values
    expandMNS(config)

    # Adding the repo and the tree to the config
    config['git_repo'] = repo
    config['git_tree'] = tree

    if cmd_exists("lmod"):
        config['module_cmd'] = "lmod"
    elif cmd_exists("modulecmd"):
        config["module_cmd"] = "modulecmd"
    else:
        sys.stdout.write("Neither modulecmd nor lmod has been found in your path. Please install either one of them to continue. (Preferably choose lmod for more functionalities)")
        exit(40)

    return config


# Generate a value for the releasedir field of the dict.
def generateReleasedir(hashTable):
    # If we build the HEAD of a branch, we have to find out which branch it is
    if hashTable['release'] == 'HEAD':
        # if a branch is provided, we use it
        if 'branch' in hashTable:
            branch = hashTable['branch']
        # if not provided, we use the one pointed by HEAD.
        else:
            branch = hashTable['git_repo'].active_branch.name

        tree = hashTable['git_repo'].heads[branch].commit.tree
        release = tree['VERSION'].data_stream.read().splitlines()[0]
        shortVersion = re.match('[0-9]*\.[0-9]*', release).group(0)
        hashTable['releasedir'] = os.path.join(branch, 'v' + shortVersion + '-' + time.strftime("%Y%m%d"))
    # If we have been given a more specific release, the branch doesn't matter and we proceed directly
    else:
        # But if we were provided a branch, if the branch really is the branch of the release, we will build in the corresponding <branch> directory
        if 'branch' in hashTable:
            branch = hashTable['branch']
            tree = hashTable['git_repo'].commit(hashTable['release']).tree
            # If the branch given isn't the same as the branch of the given release, we don't go any further
            commitBranches = subprocess.check_output(['git', 'branch', '--contains', hashTable['release']]).split("\n")
            if not any(True for line in commitBranches if re.search("[^\s]*$", line).group(0) == branch):
                sys.exit("\nThe release you want to build is not part of the branch you have given.\n")
            release = tree['VERSION'].data_stream.read().splitlines()[0]
            shortVersion = re.match('[0-9]*\.[0-9]*', release).group(0)
            hashTable['releasedir'] = os.path.join(branch, 'v' + shortVersion + '-' + time.strftime("%Y%m%d"))
        # if not provided, then we find out if we build a tag or a commit and build in the appropriate directory
        else:
            tagRegex = 'v?[0-9]*\.[0-9]*\.[0-9]*'
            branch = 'tag'
            # VERSION-{tag, SHA1}
            tree = hashTable['git_repo'].commit(hashTable['release']).tree
            release = tree['VERSION'].data_stream.read().splitlines()[0]
            if re.match(tagRegex, hashTable['release']):
                hashTable['releasedir'] = os.path.join('tag', 'v' + release)
            else:
                commit = hashTable['release'][:7]
                hashTable['releasedir'] = os.path.join('commit', 'v' + release + '-' + commit)


# Generate a value for the rootinstall field of the dict.
def generateRootinstall(hashTable):
    	hashTable['rootinstall'] = os.path.join(hashTable['apps_root'], hashTable['releasedir'])


# Get the version of EasyBuild installed in the <rootdirectory>/EasyBuild directory
def getEasyBuildVersion(rootdirectory):
    # We make sure that the path given is totally expanded.
    absrootdirectory = os.path.abspath(os.path.expandvars(rootdirectory))
    # Appending temporarily to PYTHONPATH the EasyBuild directories and vsc-base
    sys.path.insert(0, os.path.join(os.path.join(absrootdirectory, '.installRef'), 'easybuild-framework'))
    sys.path.insert(0, os.path.join(os.path.join(absrootdirectory, '.installRef'), 'easybuild-easyblocks'))
    sys.path.insert(0, os.path.join(os.path.join(absrootdirectory, '.installRef'), 'easybuild-easyconfigs'))
    sys.path.insert(0, os.path.join(os.path.join(absrootdirectory, '.installRef'), 'vsc-base'))
    
    # Importing the function that EasyBuild uses to determine its own version
    from easybuild.tools.version import this_is_easybuild
    
    # Getting the version of EasyBuild from the output message
    msg = this_is_easybuild()
    version = re.search("[0-9]*\.[0-9]*\.[0-9]*", msg).group(0)
    
    # Removing from the PYTHONPATH the EasyBuild directories (cleanup)
    sys.path.pop(0)
    sys.path.pop(0)
    sys.path.pop(0)
    sys.path.pop(0)
    
    return version


# Generate a config file for EasyBuild
def easybuildConfigfileCreator(hashTable):
    trueVersion = os.path.basename(hashTable['rootinstall'])
    path = os.path.join(hashTable['srcpath'], 'config') # <srcpath>/config
    repository = 'FileRepository'
    logfile_format = ('easylog', 'easybuild-%(name)s-%(version)s-%(date)s.%(time)s.log')

    ebdirsOnPlace = os.path.join(hashTable['rootinstall'], '.ebdirs') # <rootinstall>/.ebdirs
    sourcepathOnPlace = os.path.join(ebdirsOnPlace, 'sources') # <rootinstall>/.ebdirs/sources
    buildpathOnPlace = os.path.join(ebdirsOnPlace, 'build') # <rootinstall>/.ebdirs/build
    repositorypathOnPlace = os.path.join(ebdirsOnPlace, 'eb_repo') # <rootinstall>/.ebdirs/eb_repo

    with open(os.path.join(path, 'easybuild.cfg'), 'w') as f:
        f.write('[config]\n')
        f.write('sourcepath = ' + sourcepathOnPlace + '\n')
        f.write('buildpath = ' + buildpathOnPlace + '\n')
        f.write('repository = ' + repository + '\n')
        f.write('repositorypath = ' + repositorypathOnPlace + '\n')
        f.write('module-naming-scheme = ' + hashTable['mns'] +'\n')
        # Currenctly, this option isn't working (EasyBuild 1.15.2) so we don't use it until it is fixed.
        f.write('#logfile-format = ' + logfile_format[0] + ',' + logfile_format[1] + '\n')

    with open(os.path.join(path, 'easybuild-out-place.cfg'), 'w') as f:
        f.write('[config]\n')
        f.write('module-naming-scheme = ' + hashTable['mns'] +'\n')
        # Currenctly, this option isn't working (EasyBuild 1.15.2) so we don't use it until it is fixed.
        f.write('#logfile-format = ' + logfile_format[0] + ',' + logfile_format[1] + '\n')


# Determine the name of the EasyBuild module that should be loaded depending on the MNS.
def getEasyBuildModule(hashTable):
    if hashTable['mns'] == 'ThematicMNS':
        return "base/EasyBuild/install-" + getEasyBuildVersion(hashTable['rootinstall'])
    else:
        return 'EasyBuild/install-' + getEasyBuildVersion(hashTable['rootinstall'])

# Set EasyBuild variables to default values (for cleaninstall)
def setEasyBuildVariables(hashTable):
    ebdirsRoot = os.path.join(hashTable['rootinstall'], '.ebdirs') # <rootinstall>/.ebdirs
    if not os.path.exists(ebdirsRoot):
            os.makedirs(ebdirsRoot)

    if not "eb_sourcepath" in hashTable:
        hashTable['eb_sourcepath'] = os.path.join(ebdirsRoot, 'sources')
    if not "eb_buildpath" in hashTable:
        hashTable['eb_buildpath'] = os.path.join(ebdirsRoot, 'build')
    if not "eb_repository" in hashTable:
        hashTable['eb_repository'] = 'FileRepository'
    if not "eb_repositorypath" in hashTable:
        hashTable['eb_repositorypath'] = os.path.join(ebdirsRoot, 'eb_repo')

# Resolve MNS to be used (allow for short names of the MNS)
def expandMNS(hashTable):
    if hashTable['mns'] == 'E':
        hashTable['mns'] = 'EasyBuildMNS'
    if hashTable['mns'] == 'H':
        hashTable['mns'] = 'HierarchicalMNS'
    if hashTable['mns'] == 'T':
        hashTable['mns'] = 'ThematicMNS'

# Resolve the name of the branch to be used for alternative EasyBuild repositories.
def resolveEBbranches(hashTable):
    for repo in ['framework', 'blocks', 'configs']:
        if 'git_eb'+repo in hashTable:
            gitUrl = re.search("^[^|]*", hashTable['git_eb'+repo]).group(0)
            gitBranch = re.search("[^|]*$", hashTable['git_eb'+repo]).group(0)
            hashTable['git_eb'+repo] = gitUrl
            # We set the correct branch if any has been given (priority is given to the branch-eb* option)
            if not 'branch_eb'+repo in hashTable and gitUrl != gitBranch:
                hashTable['branch_eb'+repo] = gitBranch

def isValidRelease(hashTable):

    if "release" in hashTable:
        stackRelease = re.search("/[^/]*/?$", hashTable["rootinstall"]).group(0)[2:5]
        if hashTable["release"] != "HEAD":
            givenRelease = re.search("^v?[0-9].[0-9]", hashTable["release"]).group(0)
            givenRelease = givenRelease[1:] if givenRelease[0] == "v" else givenRelease
            if givenRelease != stackRelease:
                return False
    
    if "branch" in hashTable:
        stackBranch = re.search("^/[^/]*", re.search("/[^/]*/[^/]*/?$", hashTable["rootinstall"]).group(0)).group(0)[1:]
        if hashTable["branch"] != stackBranch:
            return False
    
    return True

#######################################################################################################################



#######################################################################################################################
# Author: Maxime Schmitt
# Mail: maxime.schmitt@telecom-bretagne.eu
# Overview: Module that take care of bootstraping EasyBuild.
#######################################################################################################################

import os
import re
import sys
import shutil
import subprocess

sys.path.append('.')
from configManager import getEasyBuildVersion

#######################################################################################################################
# Function that do the bootstrap using the other functions of this module and return the associated modulePath.
def bootstrap(hashTable):

	# We install all the necessary files of EasyBuild.
	easybuildFilesInstaller(hashTable)

	# We create the modulefile of EasyBuild.
	eb_version = getEasyBuildVersion(hashTable['rootinstall'])
	modulefileCreator(hashTable, "install-" + eb_version)

	# We create some files to source to rapidly start using the bootstrap module and the associated softwares.
	modulePath = sourcefileCreator(hashTable)

	return modulePath

#######################################################################################################################


#######################################################################################################################
# Functions that do the different important tasks of the bootstrap.

# Get and put at the right place all the EasyBuild files.
def easybuildFilesInstaller(hashTable):
	# We create the directory that we need to install the EasyBuild sources
	easyBuildDir = os.path.join(hashTable['rootinstall'], '.installRef')
	if not os.path.exists(easyBuildDir):
		os.makedirs(easyBuildDir)

	ebSources = {}
	# If we were provided alternative sources for the EasyBuild files, we create the useful variables
	if any(True for x in ['gh_ebuser', 'git_ebframework', 'git_ebblocks', 'git_ebconfigs'] if x in hashTable):
		if 'gh_ebuser' in hashTable:
			ebSources['easybuild-framework'] = ('https://github.com/' + hashTable['gh_ebuser'] + '/easybuild-framework.git', \
				hashTable['branch_ebframework'] if 'branch_ebframework' in hashTable else None)

			ebSources['easybuild-easyblocks'] = ('https://github.com/' + hashTable['gh_ebuser'] + '/easybuild-easyblocks.git', \
				hashTable['branch_ebblocks'] if 'branch_ebblocks' in hashTable else None)

			ebSources['easybuild-easyconfigs'] = ('https://github.com/' + hashTable['gh_ebuser'] + '/easybuild-easyconfigs.git', \
				hashTable['branch_ebconfigs'] if 'branch_ebconfigs' in hashTable else None)

		if 'git_ebframework' in hashTable:
			ebSources['easybuild-framework'] = (hashTable['git_ebframework'], hashTable['branch_ebframework'] if 'branch_ebframework' in hashTable else None)
	
		for v in ['blocks', 'configs']:
			if 'git_eb'+v in hashTable:
				ebSources['easybuild-easy'+v] = (hashTable['git_eb'+v], hashTable['branch_eb'+v] if 'branch_eb'+v in hashTable else None)

	# Import any alternative EasyBuild files from given sources
	for k,v in ebSources.iteritems():
		if v[1] != None:
			subprocess.check_call(['git', 'clone', v[0], '-b', v[1], os.path.join(os.path.join(hashTable['rootinstall'], '.installRef'), k)])
		else:
			subprocess.check_call(['git', 'clone', v[0], os.path.join(os.path.join(hashTable['rootinstall'], '.installRef'), k)])

	# If the backward compability is activated, we complete the sources with the ones from the subtrees.
	if 'subtrees' in hashTable and hashTable['subtrees']:
		# Complete the EasyBuild files with the ones from the subtree if necessary
		if any(True for x in ['easybuild-framework', 'easybuild-easyblocks', 'easybuild-easyconfigs'] if not x in ebSources):
			pwd = os.getcwd()
			os.chdir(hashTable['srcpath'])
	
			for k in ['easybuild-framework', 'easybuild-easyblocks', 'easybuild-easyconfigs']:
				if not k in ebSources:
					originalBranch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD'])[:-1]
					if hashTable['release'] != 'HEAD':
						subprocess.check_call(['git', 'branch', k, hashTable['release']])
						subprocess.check_call(['git', 'checkout', k])
					elif 'branch' in hashTable:
						subprocess.check_call(['git', 'branch', k, hashTable['branch']])
						subprocess.check_call(['git', 'checkout', k])
					else:
						subprocess.check_call(['git', 'branch', k])
					subprocess.check_call(['git', 'filter-branch', '-f', '--subdirectory-filter', 'easybuild/'+k, k])
					subprocess.check_call(['git', 'clone', hashTable['srcpath'], '-b', k, os.path.join(os.path.join(hashTable['rootinstall'], '.installRef'), k)])
					subprocess.check_call(['git', 'checkout', originalBranch])
					subprocess.check_call(['git', 'branch', '-D', k])
			
			os.chdir(pwd)
	# If not, we output the missing components and exit.
	else:
		if any(True for x in ['easybuild-framework', 'easybuild-easyblocks', 'easybuild-easyconfigs'] if not x in ebSources):
			for k in ['easybuild-framework', 'easybuild-easyblocks', 'easybuild-easyconfigs']:
				if not k in ebSources:
					sys.stdout.write('Missing sources for ' + k + '\n')
			sys.stdout.write('Please provide complete EasyBuild sources to allow complete installation.\n')
			exit(60)
	# Adding vsc-base from the hpcugent git repository
	subprocess.check_call(['git', 'clone', 'https://github.com/hpcugent/vsc-base.git', os.path.join(os.path.join(hashTable['rootinstall'], '.installRef'), 'vsc-base')])


# Create the EasyBuild module file and the associated symlink and put them at the right places.
# If the ThematicMNS module naming scheme is used, it is also installed.
def modulefileCreator(hashTable, moduleName):
	modulesDirPath = os.path.join(os.path.join(hashTable['rootinstall'], 'core'), 'modules')
	# Adapt the location of the modulfile to the chosen MNS
	if hashTable['mns'] == "ThematicMNS":
		# We create the directories we need to install EasyBuild
		EBmoduleDir = os.path.join('base', 'EasyBuild')
		easybuildPath = os.path.join(os.path.join(modulesDirPath, 'all'), EBmoduleDir)
		if not os.path.exists(easybuildPath):
			os.makedirs(easybuildPath)
		easybuildPath = os.path.join(os.path.join(modulesDirPath, 'base'), EBmoduleDir)
		if not os.path.exists(easybuildPath):
			os.makedirs(easybuildPath)
		# We install the ThematicMNS
		setThematicMNS(hashTable)

	else:
		# We create the directories we need to install EasyBuild
		EBmoduleDir = 'EasyBuild'
		easybuildPath = os.path.join(os.path.join(modulesDirPath, 'all'), EBmoduleDir)
		if not os.path.exists(easybuildPath):
			os.makedirs(easybuildPath)
		easybuildPath = os.path.join(os.path.join(modulesDirPath, 'base'), EBmoduleDir)
		if not os.path.exists(easybuildPath):
			os.makedirs(easybuildPath)

	# Path to the actual module file
	moduleFilePath = os.path.join(os.path.join(os.path.join(modulesDirPath, 'all'), EBmoduleDir), moduleName)

	with open(moduleFilePath, "w") as f:
		f.write("\
#%Module\n\
\n\
proc ModulesHelp { } {\n\
	puts stderr {   EasyBuild is a software build and installation framework\n\
written in Python that allows you to install software in a structured,\n\
repeatable and robust way. - Homepage: http://hpcugent.github.com/easybuild/\n\
This module provides the development version of EasyBuild.\n\
}\n\
}\n\
module-whatis {EasyBuild is a software build and installation framework\n\
written in Python that allows you to install software in a structured,\n\
repeatable and robust way. - Homepage: http://hpcugent.github.com/easybuild/\n\
This module provides the development version of EasyBuild.\n\
}\n\
set root    " + os.path.join(hashTable['rootinstall'], '.installRef') + "\n\
conflict    EasyBuild\n\
prepend-path    PATH            \"$root/easybuild-framework\"\n" \
+ ("\nprepend-path    PYTHONPATH      \"$root/MNS\"\n" if hashTable['mns'] == 'ThematicMNS' else "\n") + \
"prepend-path    PYTHONPATH      \"$root/easybuild-framework\"\n\
prepend-path    PYTHONPATH      \"$root/easybuild-easyblocks\"\n\
prepend-path    PYTHONPATH      \"$root/easybuild-easyconfigs\"\n\
prepend-path    PYTHONPATH      \"$root/vsc-base/lib\"\n\
")
	
	# Path to the symlink to the module file.
	symlinkPath = os.path.join(os.path.join(os.path.join(modulesDirPath, 'base'), EBmoduleDir), moduleName)
	os.symlink(moduleFilePath, symlinkPath)


# Install the ThematicMNS module naming scheme.
def setThematicMNS(hashTable):
	mnsroot = os.path.join(os.path.join('.installRef', 'MNS'), 'easybuild')
	
	# Extends the Python path in such a way that we included this architecture in the easybuild namespace.
	mnsPath = os.path.join(hashTable['rootinstall'], os.path.join(os.path.join(mnsroot, 'tools'), 'module_naming_scheme'))
	if not os.path.exists(mnsPath):
		os.makedirs(mnsPath)
	with open(os.path.join(mnsPath, '__init__.py'), "w") as f:
		f.write("\
from pkgutil import extend_path\n\
__path__ = extend_path(__path__, __name__)\n\
")
	shutil.copyfile(os.path.join(mnsPath, '__init__.py'), \
		os.path.join(os.path.join(os.path.join(hashTable['rootinstall'] ,mnsroot), 'tools'), '__init__.py'))
	shutil.copyfile(os.path.join(mnsPath, '__init__.py'), \
		os.path.join(os.path.join(hashTable['rootinstall'], mnsroot), '__init__.py'))

	# Import the module naming scheme
	thematicStr = hashTable['git_tree']['bin/ThematicMNS.py'].data_stream.read()
	with open(os.path.join(mnsPath, 'ThematicMNS.py'), 'w') as f:
		f.write(thematicStr)


# Create some files to source to easily start using the EasyBuild module and the associated softwares.
# These files are put in the <rootinstall> directory.
def sourcefileCreator(hashTable):
	trueVersion = os.path.basename(hashTable['rootinstall'])
	# Variable containing all the required path for the MODULEPATH
	modulePath = ""
	moduleClasses = ['bio', 'cae', 'chem', 'compiler', 'data', 'debugger', 'devel', 'geo', 'lang', 'lib', 'math', \
	'mpi', 'numlib', 'perf', 'phys', 'system', 'toolchain', 'tools', 'vis', 'base']
	for moduleclass in moduleClasses:
		modulePath += os.path.join(os.path.join(os.path.join('$RESIF_ROOTINSTALL', 'core'), 'modules'), moduleclass) + ":"

	prependModulepath = (hashTable['prepend_modulepath'] + ":") if 'prepend_modulepath' in hashTable else ""
	appendModulepath = hashTable['append_modulepath'] if 'append_modulepath' in hashTable else ""
	modulePath = prependModulepath + modulePath + appendModulepath

	# We create the files to source to use the infrastructure.
	# The admin file is there to easily add software in the ulhpc swset without any manual changes to the config.
	with open(os.path.join(hashTable['rootinstall'], "LOADME-" + trueVersion + ".sh"), "w") as f:
		modulevar = "export EASYBUILD_MODULES_TOOL=Lmod" if hashTable["module_cmd"] == "lmod" else "#export EASYBUILD_MODULES_TOOL=Lmod"
		f.write("\
export RESIF_ROOTINSTALL=" + hashTable['rootinstall'] + "\n\
export MODULEPATH=" + modulePath + "\n\
" + modulevar + "\n\
export EASYBUILD_MODULE_NAMING_SCHEME=" + hashTable['mns'] + "\n\
export RESIF_OUT_PLACE=False\n\
")
	# The user file is there to easily add software locally without any manual change to the config.
	with open(os.path.join(hashTable['rootinstall'], "LOADME-" + trueVersion + "-out-place.sh"), "w") as f:
		modulevar = "export EASYBUILD_MODULES_TOOL=Lmod" if hashTable["module_cmd"] == "lmod" else "#export EASYBUILD_MODULES_TOOL=Lmod"
		f.write("\
export RESIF_ROOTINSTALL=" + hashTable['rootinstall'] + "\n\
export MODULEPATH=" + modulePath + "\n\
" + modulevar + "\n\
export EASYBUILD_MODULE_NAMING_SCHEME=" + hashTable['mns'] + "\n\
export RESIF_OUT_PLACE=True\n\
")
	
	return modulePath.replace('$RESIF_ROOTINSTALL', hashTable['rootinstall'])

#######################################################################################################################


#######################################################################################################################
# Author: Maxime Schmitt
# Mail: maxime.schmitt@telecom-bretagne.eu
# Overview: Module that take care of installing a given software set using a given EasyBuild module.
#######################################################################################################################

import os
import sys
import subprocess
import yaml
import time
import re
import glob

#######################################################################################################################

# Installs the software sets.
def build(hashTable):
	# Assume that the MODULEPATH is already set to a correct value
    easybuild = hashTable['easybuild_module']
    
    if "swsets_config" in hashTable:
        stream = file(hashTable['swsets_config'], 'r')
    else:
        stream = hashTable["git_tree"]['config/swsets.yaml'].data_stream.read()

    swsets = yaml.load(stream)

    # We define the options that are going to be passed to EasyBuild
    sharedOptions = defineSharedOptions(hashTable)

    # We use the additional options if any
    if 'eb_options' in hashTable:
        additionalOptions = " " + hashTable["eb_options"]
    else:
        additionalOptions = ""

    for swset in hashTable['swsets']:
        # We set the installpath. If there is no installpath given, we stop the execution.
        installpath = setInstallpath(hashTable, swset)

        # We add the place where the software will be installed to the MODULEPATH for the duration of the installation
        # so that EasyBuild will not instantly forget that it has installed them after it is done (problematic for dependency resolution)
        # Part for environment-modules (come later for Lmod)
        # We also add the core software set to ensure that EasyBuild will be available
        corePath = os.path.join(hashTable['rootinstall'], 'core', 'modules', 'all')
        if hashTable["module_cmd"] == "modulecmd":
            try:
                os.environ['MODULEPATH'] = ':'.join([os.environ['MODULEPATH'], os.path.join(os.path.join(installpath[15:], 'modules'), 'all'), corePath])
            except KeyError:
                os.environ['MODULEPATH'] = ':'.join([os.path.join(os.path.join(installpath, 'modules'), 'all'), corePath])
    	
    	process = subprocess.Popen('/bin/bash', stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        # If we're doing a cleaninstall, we don't want the EASYBUILD_INSTALLPATH environment variable to interfere.
        if 'iscleaninstall' in hashTable and hashTable['iscleaninstall']:
            process.stdin.write('unset EASYBUILD_INSTALLPATH\n')

        # Lmod part for MODULEPATH management
        if hashTable["module_cmd"] == "lmod":
            process.stdin.write("module use " + corePath + "\n")
            process.stdin.write("module use " + os.path.join(os.path.join(installpath[15:], 'modules'), 'all') + "\n")

    	process.stdin.write('module load ' + easybuild + '\n')
    	
        alreadyInstalled = False
    	# If it actually exist in the yaml file, we install the listed software.
    	if swset in swsets:
            swsetStart= time.time()
    	    for software in swsets[swset]:
                sys.stdout.write("Now starting to install " + software[:-3] + "\n")
    	        process.stdin.write('eb ' + software + installpath + sharedOptions + additionalOptions + ' --robot\n')
    	        # Command to have at the end of the output the execution code of the last command
    	        process.stdin.write('echo $?\n')
    	        out = ""
    	        while True:
    	            out = process.stdout.readline()
                    if re.search("\(module found\)", out) != None:
                        alreadyInstalled = True
    	            try:
    	                i = int(out)
    	            except ValueError:
    	                i = -1
    	            if i < 0:
    	                sys.stdout.write(out)
    	            else:
    	                if i == 0:
                            if alreadyInstalled:
                                sys.stdout.write(software[:-3] + " was already installed. Nothing to be done.\n")
                                alreadyInstalled = False
                            else:
                                sys.stdout.write('Successfully installed ' + software[:-3] + '.\n')
    	                else:
    	                    sys.stdout.write('Failed to install ' + software[:-3] + '\n' + 'Operation failed with return code ' + out + '\n')
    	                    exit(out)
    	                break
            if swset != 'core':
                if not 'out_place' in hashTable or not hashTable['out_place']:
                    swsetModulefileCreator(hashTable, installpath[15:], swset)
            swsetEnd = time.time()
            swsetDuration = swsetEnd - swsetStart
            m, s = divmod(swsetDuration, 60)
            h, m = divmod(m, 60)
            swsetDurationStr = "%d:%d:%d" % (h, m, s)
    	    sys.stdout.write("Software set " + swset + " Successfully installed. Build duration: " + swsetDurationStr + ".\n")
    	# If it doesn't, we print an error message as well as an help to use the script.
    	else:
    	    sys.stdout.write("Error: Invalid set of software.\nYou asked for the software set named " + swset  + \
    	    	" to be installed but it wasn't found in your software set configuration file.\n")
    	    sys.stdout.write("The configuration file location is: " + os.path.abspath(hashTable['swsets_config']) + \
    	    	" and the software sets found in it are:\n")
    	    for k in swsets.iteritems():
    	        sys.stdout.write("- " + k[0] + "\n")
    	    exit(20)


# Defines options to pass to EasyBuild and return them in a string.
def defineSharedOptions(hashTable):
    options = ""

    if 'eb_sourcepath' in hashTable:
        options += ' --sourcepath=' + os.path.abspath(os.path.expandvars(hashTable['eb_sourcepath']))
    if 'eb_buildpath' in hashTable:
        options += ' --buildpath=' + os.path.abspath(os.path.expandvars(hashTable['eb_buildpath']))
    if 'eb_repository' in hashTable:
        options += ' --repository=' + hashTable['eb_repository']
    if 'eb_repositorypath' in hashTable:
        options += ' --repositorypath=' + os.path.abspath(os.path.expandvars(hashTable['eb_repositorypath']))
    if 'mns' in hashTable:
        options += ' --module-naming-scheme=' + hashTable['mns']

    # If this file doesn't exist, it wont crash, EasyBuild will just ignore it and use other sources for the 
    # options (default config file and so on).
    if 'out_place' in hashTable and hashTable['out_place']:
        easybuild_config = os.path.join(os.path.join(hashTable['srcpath'], 'config'), 'easybuild-out-place.cfg')
    else:
        easybuild_config = os.path.join(os.path.join(hashTable['srcpath'], 'config'), 'easybuild.cfg')
    options += ' --configfile=' + easybuild_config

    return options


# Return the installpath for the given swset.
def setInstallpath(hashTable, swset):
    installpath = ""
    # An installpath has to be provided, if not explicitely specified we use the rootinstall.
    if 'installdir' in hashTable:
        installpath = ' --installpath=' + os.path.join(hashTable['installdir'], swset)
    elif 'out_place' in hashTable and hashTable['out_place']:
        return installpath
    elif 'rootinstall' in hashTable:
            installpath = ' --installpath=' + os.path.join(hashTable['rootinstall'], swset)
    else:
        sys.stdout.write("\
No information on where to install the softwares has been found. Please provide this information.\n\
To do so, either:\n\
- use the --rootinstall option to define the root directory of the EasyBuild install.\n\
- set the RESIF_ROOTINSTALL environement variable to define the same information.\n\
- set the rootinstall field in your configuration file if you are providing one. (using the --configfile option)\n\
- use the --installdir option to define an alternative place to install (modules will be installed in <eb-installpath>/<swset>/modules)\n\
- set the RESIF_INSTALLDIR environement variable to define the same information.\n\
- set the installdir field in your configuration file if you are providing one. (using the --configfile option)\n\
- set --out-place to True to install use the EasyBuild default installpath\n\
- set RESIF_OUT_PLACE to True to define the same information\n\
")
        exit(10)

    return installpath

# For a given logfile we get the name of the software, the start and end dates of the build and convert them to timestamp and then determine the duration of the build.
def getSoftwareBuildTimes(logfile):
    logfileReduced = re.search("[^/]*$", logfile).group(0)
    software = re.findall("-.*-", logfileReduced)[0][1:-1]
    with open(logfile, "r") as log:
        raw = log.readlines()
        stime = "%s %s"%(raw[0].split()[1],raw[0].split()[2])
        stime = time.mktime(time.strptime(stime[:-4], "%Y-%m-%d %H:%M:%S"))
        etime = "%s %s"%(raw[-1].split()[1],raw[-1].split()[2])
        etime = time.mktime(time.strptime(etime[:-4], "%Y-%m-%d %H:%M:%S"))
        softwareDuration = etime - stime
        
    return (software, softwareDuration)

# Create a module file and the associated symlink (if not already existing) to load the software set and put them at the right places.
def swsetModulefileCreator(hashTable, installpath, moduleName):
    if 'installdir' in hashTable:
        modulesDirPath = os.path.join(os.path.join(hashTable['installdir'], 'core'), 'modules')
    else:
        modulesDirPath = os.path.join(os.path.join(hashTable['rootinstall'], 'core'), 'modules')
    # Adapt the location of the modulfile to the chosen MNS
    if hashTable['mns'] == "ThematicMNS":
        # We create the directories we need to install EasyBuild
        EBmoduleDir = os.path.join('base', 'swsets')
        easybuildPath = os.path.join(os.path.join(modulesDirPath, 'all'), EBmoduleDir)
        if not os.path.exists(easybuildPath):
            os.makedirs(easybuildPath)
        easybuildPath = os.path.join(os.path.join(modulesDirPath, 'base'), EBmoduleDir)
        if not os.path.exists(easybuildPath):
            os.makedirs(easybuildPath)

    else:
        # We create the directories we need to install EasyBuild
        EBmoduleDir = 'swsets'
        easybuildPath = os.path.join(os.path.join(modulesDirPath, 'all'), EBmoduleDir)
        if not os.path.exists(easybuildPath):
            os.makedirs(easybuildPath)
        easybuildPath = os.path.join(os.path.join(modulesDirPath, 'base'), EBmoduleDir)
        if not os.path.exists(easybuildPath):
            os.makedirs(easybuildPath)

    # Path to the actual module file
    moduleFilePath = os.path.join(os.path.join(os.path.join(modulesDirPath, 'all'), EBmoduleDir), moduleName)

    with open(moduleFilePath, "w") as f:
        f.write("\
#%Module\n\
\n\
proc ModulesHelp { } {\n\
    puts stderr {   " + moduleName + " software set.\n\
}\n\
}\n\
module-whatis {" + moduleName + " software set.\n\
}\n\
set root        " + installpath + "\n\
prepend-path    MODULEPATH      \"$root/modules/all\"\n\
")
    
    # Path to the symlink to the module file.
    symlinkPath = os.path.join(os.path.join(os.path.join(modulesDirPath, 'base'), EBmoduleDir), moduleName)
    if not os.path.exists(symlinkPath):
        os.symlink(moduleFilePath, symlinkPath)

#######################################################################################################################



from setuptools import setup

setup(
    name='resif',
    version='2.0.0',
    author='ULHPC',
    author_email='hpc-sysadmins@uni.lu',
    maintainer='ULHPC',
    maintainer_email='hpc-sysadmins@uni.lu',
    url='https://hpc.uni.lu/',
    license='GPLv2',
    description='Command line interface to deploy an EasyBuild infrastructure and manage it.',
    long_description='This module has been developped by the University of Luxembourg in order to improve the management of its clusters. This application will help you easily deploy an EasyBuild installation and deploy software on it, in a heavily automated way.',
    classifiers=[
    'License :: OSI Approved',
    'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Operating System :: Unix',
    'Operating System :: MacOS',
    'Operating System :: POSIX',
    'Operating System :: POSIX :: Linux',
    'Topic :: Scientific/Engineering',
    'Topic :: System :: Installation/Setup',
    'Topic :: System :: Systems Administration',
    'Topic :: Utilities',
    'Natural Language :: English',
    'Intended Audience :: Science/Research',
    'Intended Audience :: System Administrators',
    'Environment :: Console',
    ],
    packages=['resif'],
    install_requires=[
        'click >= 3.3',
        'GitPython >= 0.3',
        'PyYaml >= 3.10',
    ],
    entry_points='''
        [console_scripts]
        resif=resif.resif:resif
    ''',
)
